from __future__ import annotations

import kenlm
import os
import sentencepiece as spm
import torch
from transformers import AutoTokenizer, AutoModelForCausalLM
import warnings


class Reranker:
    """
    Class for GEC which can use various underlying models. The class behaves as a singleton object and is intended
    to have 1 "active" preloaded model, which is selected using `Reranker.set_current_model` method.

    Ranking can be performed using either the entire sentence or only correction span via `Reranker.rank` method.

    The list of supported models is given in the `Reranker.Models` subclass.
    """

    class Models:
        kenlm_en = "kenlm/en"
        distillgpt2 = "distilgpt2"

    model_name_: str = Models.kenlm_en
    models_: dict[str, object] = dict()
    tokenizers_: dict[str, object] = dict()

    def __init__(self, model_name: str | None = None):
        Reranker.set_current_model(model_name or Reranker.Models.kenlm_en)

    @staticmethod
    def set_current_model(model_name: str):
        """
        Select model which will be used for reranking.

        If selected model can use CUDA device, then it will be automatically sent onto the device.

        Unloads models that were previously loaded / selected.

        Parameters
        ----------
        model_name : str
            Model name available in the `Reranker.Models`.
        """
        if model_name not in Reranker.models_:

            if model_name == Reranker.Models.kenlm_en:
                if not os.path.exists("lm/en.arpa.bin"):
                    raise RuntimeError("The KenLM model is not found. Download the KenLM model and tokenizer first (eg from: https://huggingface.co/edugp/kenlm)")

                Reranker.models_[Reranker.Models.kenlm_en] = kenlm.LanguageModel("lm/en.arpa.bin")
                Reranker.tokenizers_[Reranker.Models.kenlm_en] = spm.SentencePieceProcessor(model_file="lm/en.sp.model")

            elif model_name == Reranker.Models.distillgpt2:
                device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
                model = AutoModelForCausalLM.from_pretrained(model_name)
                model.to(device)
                Reranker.models_[model_name] = model

                tokenizer = AutoTokenizer.from_pretrained(model_name)
                tokenizer.pad_token = tokenizer.eos_token
                Reranker.tokenizers_[model_name] = tokenizer

            else:
                raise NotImplementedError(f"Unknown model name {model_name}")

        Reranker.model_name_ = model_name
        Reranker.unload_unused_models()

    @staticmethod
    def unload_unused_models():
        """
        Unloads all models and their tokenizers except for the model in `model_name_`.
        """
        cur_models = list(Reranker.models_)
        for model_name in cur_models:
            if model_name == Reranker.model_name_:
                continue
            Reranker.unload_model(model_name)

    @staticmethod
    def unload_model(model_name: str) -> bool:
        """
        Unload selected model and its tokenizer.

        Returns
        -------
        bool
            `True` if the model was unloaded, `False` otherwise (eg if model was not previously loaded).
        """
        if model_name in Reranker.models_:
            del Reranker.models_[model_name]
            del Reranker.tokenizers_[model_name]
            return True
        return False

    @staticmethod
    def rank(context: str, original: str, replacements: list[str], full_sentence_score: bool = False) -> list[tuple[str, float]]:
        """
        Rank `replacements` of the `original` text in the given `context` using log-likelihood score.

        Ranking is performed using model `Reranker.model_name_` (see `Reranker.set_current_model`).

        Parameters
        ----------
        context : str
            Context, inside which the original text is defined, and where the replacement happens.
        original : str
            Original part of the context that needs to be ranked agains the `replacements`.
        replacements : list[str]
            Suggested replacements for `original`.
        full_sentence_score : bool, default=False
            Indicates whether scoring should be performed on the full sentence (True) or only on the
            correction span (False), that includes context before the original and the replacement. The former is
            more accurate, but more computationally expensive.

        Return
        ------
        list[tuple[str, float]]
            List of replacements and their log-likelihood scores. The first item always corresponds to `original`.
            The scores for replacements are listed in the same order as items in the `replacements`.
        """
        if Reranker.model_name_ not in Reranker.models_:  # load on demand
            Reranker.set_current_model(Reranker.model_name_)

        method = {
            Reranker.Models.kenlm_en: Reranker.rank_kenlm_en,
            Reranker.Models.distillgpt2: Reranker.rank_distillgpt2,
        }
        kwargs = dict(
            context=context,
            original=original,
            replacements=replacements,
            full_sentence_score=full_sentence_score
        )
        res = method[Reranker.model_name_](**kwargs)
        sorted_res = sorted(res, key=lambda k: k[1], reverse=True)
        return sorted_res

    @staticmethod
    def prepare_candidates(context: str, original: str, replacements: list[str], full_context: bool = False) -> list[str]:
        """
        Prepare candidate phrases for ranking.

        Note:
        * if `original` is encountered several times in the `context`, then only the **1st** occurrence
          will be replaced to generate candidates.

        Parameters
        ----------
        context: str
            Context for evaluation of ranking.
        original: str
            Original part of the context.
        replacements: list[str]
            Replacements for the `original`, that will be evaluated for ranking.
        full_context: bool, default=False
            Indicates whether candidates should use full context (True) or only part of the context before
            the `original` (False).
        """

        count = context.count(original)
        if count == 0:
            raise ValueError(f"Original text '{original}' not found in context '{context}'. Ensure the original text exists exactly as specified in the context.")

        if count > 1:
            warnings.warn(f"Original '{original}' is present in context multiple times. Only the first occurrence will be replaced when generating candidates.", UserWarning, stacklevel=2)

        subctx = None
        if not full_context:
            subctx = context.split(original)[0]
            res = [subctx + original]
        else:
            res = [context]

        for replacement in replacements:
            if full_context:
                res.append(context.replace(original, replacement, 1))
            else:
                res.append(subctx + replacement)
        return res

    @staticmethod
    def rank_kenlm_en(context: str, original: str, replacements: list[str], full_sentence_score: bool = False) -> list[tuple[str, float]]:
        tokenizer: spm.SentencePieceProcessor = Reranker.tokenizers_[Reranker.Models.kenlm_en]
        model: kenlm.LanguageModel = Reranker.models_[Reranker.Models.kenlm_en]

        res = []
        all_replacements = [original] + replacements
        candidates = Reranker.prepare_candidates(context, original, replacements, full_context=full_sentence_score)
        for idx, candidate in enumerate(candidates):
            tokens = tokenizer.encode(candidate, out_type=str)
            cur_score = model.score(" ".join(tokens), bos=True, eos=True)
            res.append((all_replacements[idx], cur_score))

        return res

    @staticmethod
    def rank_distillgpt2(context: str, original: str, replacements: list[str], full_sentence_score: bool = False) -> list[tuple[str, float]]:
        model = Reranker.models_[Reranker.Models.distillgpt2]
        device = next(model.parameters()).device

        candidates = Reranker.prepare_candidates(context, original, replacements, full_context=full_sentence_score)
        tokenizer = Reranker.tokenizers_[Reranker.Models.distillgpt2]
        tokens = tokenizer(candidates, return_tensors="pt", padding=True).to(device)
        with torch.no_grad():
            outputs = model(**tokens)

        # shift for causal LM
        shift_logits = outputs.logits[:, :-1]
        shift_labels = tokens.input_ids[:, 1:]
        log_probs = torch.nn.functional.log_softmax(shift_logits, dim=-1)
        token_log_probs = log_probs.gather(-1, shift_labels.unsqueeze(-1)).squeeze(-1)

        shift_mask = tokens.attention_mask[:, 1:]
        likelihoods = (shift_mask * token_log_probs).sum(-1) / shift_mask.sum(-1)

        all_replacements = [original] + replacements
        return list(zip(all_replacements, likelihoods.tolist()))



from functools import partial
import os
import pandas as pd
import re
import socket
import toml

from .entities import GrammarDescription, VocabularyPlaceholders, VocabularyInfo
from .registry import GrammarRegistry
from .vocabulary import Vocabulary


def register_miniphrase():
    cur_dir = os.path.split(__file__)[0]
    miniphrase_dir = os.path.join(cur_dir, "miniphrase")
    register_grammar_topics(miniphrase_dir)


def register_grammar_topics(grammar_dir: str | None = None):
    """
    Read and register descriptions of grammar exercises.
    """

    cur_dir = os.path.split(__file__)[0]
    default_grammar_dir = os.path.join(cur_dir, "topics")
    grammar_dir = grammar_dir or default_grammar_dir

    # iterate over files in `grammar_dir`, load descriptions of topics and exercises and register them.
    for file in sorted(os.listdir(grammar_dir)):
        if not file.endswith(".toml"):
            continue
        with open(os.path.join(grammar_dir, file), "r") as f:
            grammar_descr_dict = toml.loads(f.read())
            grammar_descr = GrammarDescription.from_dict(grammar_descr_dict)
            GrammarRegistry.register_grammar_topic(grammar_descr)


def get_vocabulary_names() -> list[str]:
    """
    Returns list with names of the built-in vocabularies.
    """
    vocabs_dir = os.path.join(os.path.dirname(__file__), "vocabs")
    res = []
    for file in sorted(os.listdir(vocabs_dir)):
        with open(os.path.join(vocabs_dir, file), "r") as f:
            vocab_info = toml.loads(f.read())
        res.append(vocab_info["name"])
    return res


def get_vocabularies_info() -> list[VocabularyInfo]:
    """
    Returns list with names of the built-in vocabularies.
    """
    vocabs_dir = os.path.join(os.path.dirname(__file__), "vocabs")
    res = []
    for file in sorted(os.listdir(vocabs_dir)):
        vocab = Vocabulary(os.path.join(vocabs_dir, file))
        res.append(
            VocabularyInfo(
                name=vocab.name,
                num_words=len(vocab.nouns) + len(vocab.verbs) + len(vocab.adjectives) + len(vocab.adverbs),
                num_nouns=len(vocab.nouns),
                num_verbs=len(vocab.verbs),
                num_adjectives=len(vocab.adjectives),
                num_adverbs=len(vocab.adverbs),
            )
        )
    return res


def apply_vocabulary(s: str, vocab: Vocabulary) -> str:
    """
    Takes input string and replaces placeholders with respective minilex entities.
    """

    vocab_function = {
        VocabularyPlaceholders.Verb: vocab.random_verb,
        VocabularyPlaceholders.VerbSingular3rd: vocab.random_verb_singular_3rd,
        VocabularyPlaceholders.VerbPast: vocab.random_verb_past,
        VocabularyPlaceholders.VerbProgressive: vocab.random_verb_progressive,
        VocabularyPlaceholders.Noun: vocab.random_noun,
        VocabularyPlaceholders.ANoun: vocab.random_anoun,
        VocabularyPlaceholders.NounPlural: vocab.random_noun_plural,
        VocabularyPlaceholders.NounNonPerson: vocab.random_noun_non_person,
        VocabularyPlaceholders.ANounNonPerson: vocab.random_anoun_non_person,
        VocabularyPlaceholders.Person: vocab.random_person,
        VocabularyPlaceholders.Number12: partial(vocab.random_number, max=12),
        VocabularyPlaceholders.Number60: partial(vocab.random_number, max=60),
        VocabularyPlaceholders.Number100: partial(vocab.random_number, max=100),
        VocabularyPlaceholders.Number1000: partial(vocab.random_number, max=1000),
        VocabularyPlaceholders.Number100k: partial(vocab.random_number, max=100_000),
        VocabularyPlaceholders.LargeNumber: vocab.random_number_large,
        VocabularyPlaceholders.Weekday: partial(vocab.random_weekday),
        VocabularyPlaceholders.Season: partial(vocab.random_season),
        VocabularyPlaceholders.Month: partial(vocab.random_month),
        VocabularyPlaceholders.Adjective: partial(vocab.random_adjective),
        VocabularyPlaceholders.Adverb: partial(vocab.random_adverb),
        VocabularyPlaceholders.Country: partial(vocab.random_country),
        VocabularyPlaceholders.Place: partial(vocab.random_place),
        VocabularyPlaceholders.City: partial(vocab.random_city),
        VocabularyPlaceholders.Nationality: partial(vocab.random_nationality),
    }

    placeholder_patterns = re.compile("|".join(list(vocab_function)))
    for match in placeholder_patterns.findall(s):
        s = placeholder_patterns.sub(
            vocab_function[match](), string=s, count=1
        )

    return s


def split_vocabulary_top_n_words(input_vocab_file: str, top_n: int):
    """
    Splits input vocabulary into two non-overlapping vocabularies.
    The first contains top-N words having the highest usage frequency.
    And the second contains remaining words.

    Requires file with frequencies of words from here:
    * http://norvig.com/ngrams/count_1w100k.txt

    Args
    ----
    input_vocab_file: str
    top_n: int
        Keep top_n most frequent words.
    """

    input_vocab = Vocabulary(input_vocab_file)
    freq_df = pd.read_table("count_1w100k.txt", sep="\t", header=None)
    freq_info = freq_df.set_index(0).to_dict()[1]

    input_words = set(input_vocab.nouns + input_vocab.adjectives + input_vocab.adverbs + input_vocab.verbs)
    input_words_freq = [freq_info.get(word.upper(), 0) for word in input_words]
    filtered_words = pd.Series(index=input_words, data=input_words_freq).sort_values(ascending=False)[:top_n]

    res_vocab = Vocabulary()
    res_vocab.nouns = [word for word in input_vocab.nouns if word in filtered_words.index]
    res_vocab.adverbs = [word for word in input_vocab.adverbs if word in filtered_words.index]
    res_vocab.adjectives = [word for word in input_vocab.adjectives if word in filtered_words.index]
    res_vocab.verbs = [word for word in input_vocab.verbs if word in filtered_words.index]

    out_path = "out_vocab_incl.toml"
    vocab_dict = {
        "name": f"Included top-{top_n} words",
        "nouns": [word for word in input_vocab.nouns if word in filtered_words.index],
        "adverbs": [word for word in input_vocab.adverbs if word in filtered_words.index],
        "adjectives": [word for word in input_vocab.adjectives if word in filtered_words.index],
        "verbs": [word for word in input_vocab.verbs if word in filtered_words.index],
    }
    with open(out_path, "w") as f:
        toml.dump(vocab_dict, f)

    out_path = "out_vocab_excl.toml"
    vocab_dict = {
        "name": "Excluded words",
        "nouns": [word for word in input_vocab.nouns if word not in filtered_words.index],
        "adverbs": [word for word in input_vocab.adverbs if word not in filtered_words.index],
        "adjectives": [word for word in input_vocab.adjectives if word not in filtered_words.index],
        "verbs": [word for word in input_vocab.verbs if word not in filtered_words.index],
    }
    with open(out_path, "w") as f:
        toml.dump(vocab_dict, f)


def find_sublist_index(full: list[str], sublist: list[str]) -> int:
    """
    Takes on input two lists and returns index in the first list from which the second list can be found.
    If second list is not part/sublist of the first list, then return -1.
    """
    if not sublist or len(sublist) > len(full):
        return -1
    for k in range(len(full) - len(sublist) + 1):
        if full[k:k + len(sublist)] == sublist:
            return k
    return -1


def is_port_in_use(port: int, host: str = "127.0.0.1") -> bool:
    """
    Checks whether the given port is in use.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind((host, port))
            return False  # Port is free
        except OSError:
            return True  # Port is in use



"""
Factory for generation of exercises for the given topic.
"""

from __future__ import annotations

from .entities import GrammarDescription


class GrammarRegistry:
    topics: dict[str, GrammarDescription] = dict()

    @staticmethod
    def reset():
        GrammarRegistry.topics = dict()

    @staticmethod
    def register_grammar_topic(grammar_descr: GrammarDescription):
        if grammar_descr.name in GrammarRegistry.topics:
            return
        GrammarRegistry.topics[grammar_descr.name] = grammar_descr

    @staticmethod
    def get_registered_topics() -> list[str]:
        """
        Returns names of registered topics.
        """
        return list(GrammarRegistry.topics)

    @staticmethod
    def get_registered_levels() -> dict[str, list[int]]:
        """
        Returns information about registered levels for topics.
        The levels are grouped by the resource and sorted in the ascending order.
        """
        res = dict()
        for topic in GrammarRegistry.topics.values():
            for level_descr in topic.levels:
                if level_descr.resource_name not in res:
                    res[level_descr.resource_name] = list()
                if level_descr.level:
                    res[level_descr.resource_name].append(level_descr.level)

        for key in res:
            res[key] = sorted(set(res[key]))
        return res

    @staticmethod
    def find_topics(topic_name: str) -> list[str]:
        """
        Find registered topic that fully or partially matches the provided name.

        Args
        ----
        topic_name: str,
            Full or partial name of the topic to find.

        Returns
        -------
        list[str]
            List of found registered topics. If nothing is found then empty list is returned.
        """

        if not topic_name:
            return []

        res = []
        lower_topic_name = topic_name.lower()
        for cur_topic in GrammarRegistry.topics:
            if lower_topic_name in cur_topic.lower():
                res.append(cur_topic)
        return res

    @staticmethod
    def generate_answers(exercises: list[str], grammar_topic: str | None = None) -> list[list[str]]:
        """
        Generate list of possible answers for the given exercises.
        Each exercise can have several possible answers.
        """
        ...



from __future__ import annotations

import atexit
import language_tool_python as ltp
import requests
import spacy

from .reranker import Reranker
from .usecases import is_port_in_use
from .vocabulary import Vocabulary


# In the cases when suggested replacement contains "doesn't", "haven't" etc, it is broken in 2 tokens,
#   one of which is "n't", which: (1) does not belong to the vocabulary as it is not a word;
#   (2) not required for lemmatization anyways, and thus can be ignored.
IGNORE_LEMMATIZATION_TOKENS = ["n't", "n\u2019t", "\u2019t"]


####################################################
# Init NLP
try:
    nlp = spacy.load("en_core_web_sm")
except OSError:
    from spacy.cli import download
    download("en_core_web_sm")
    nlp = spacy.load("en_core_web_sm")


####################################################
# Init local Language Tool server
def get_language_tool():
    """
    Get or create `LanguageTool` instance.
    """

    def is_lt_server_running(port: int):
        LT_URL = f"http://localhost:{port}/v2/check"

        try:
            args = "text=foo&language=en"
            response = requests.post(f"{LT_URL}?{args}", timeout=2)
            return response.status_code == 200
        except requests.RequestException:
            return False

    # Note: the code below inherently assumes that LT is trying to use ports starting from `LanguageTool._MIN_PORT`
    ports = [ltp.LanguageTool._MIN_PORT + k for k in range(10)]
    for port in ports:
        if is_port_in_use(port):
            if is_lt_server_running(port):
                return ltp.LanguageTool('en-US', remote_server=f"localhost:{port}")
            # Port is in use but not running LanguageTool, try next port
            continue

        # Port is free, try to start a new LanguageTool server here.
        return ltp.LanguageTool('en-US', host='localhost')
    raise RuntimeError(f"All ports are in use by other services (tested ports: {ports})")


language_tool = get_language_tool()


####################################################
# Cleanup
def on_exit():
    if "language_tool" in globals():
        language_tool.close()

atexit.register(on_exit)


####################################################

class GrammarChecker:
    @staticmethod
    def correct_phrases(phrases: list[str], vocab: Vocabulary) -> list[str]:
        """
        Checks grammar and spelling in the provided list of phrases.
        The correction is performed wrt to the provided vocabulary, so that only the vocabulary words can appear in
            the corrected phrase.

        Returns a list of fixed phrases.
        """

        res = []
        for phrase in phrases:
            matches = language_tool.check(phrase)
            matches = fix_and_rank_matches(matches, vocab)
            res.append(ltp.utils.correct(phrase, matches))

        return res


def fix_and_rank_matches(matches: list[ltp.Match], vocab: Vocabulary, reranker_model: str | None = None) -> list[ltp.Match]:
    """
    Filters out suggested replacements that are not in the provided vocabulary.

    Filters replacements for each match so only words present in `vocab` remain, then ranks the
    remaining replacements using the specified reranker model. The function mutates the input
    `matches` in-place (each match's `replacements` is updated).

    Parameters
    ----------
    matches : list[ltp.Match]
        List of matches detected by the language tool.
    vocab : Vocabulary
        Vocabulary object used to filter valid replacements.
    reranker_model : str | None, optional
        Name of the reranker model to use for ranking replacements. Defaults to None
        (uses `Reranker.model_name_`).

    Returns
    -------
    list[ltp.Match]
        The list of matches with their replacements filtered to only include vocabulary words,
        and ranked according to the reranker model.

    Notes
    -----
    - Ranks the remaining replacements using the specified reranker model.
    """

    reranker_model = reranker_model or Reranker.model_name_
    Reranker.set_current_model(reranker_model)

    for match in matches:
        cur_replacements = filter_replacements(match.matchedText, match.replacements, vocab)
        ranked_replacements = Reranker.rank(context=match.context, original=match.matchedText, replacements=cur_replacements)
        match.replacements = [replacement for replacement, score in ranked_replacements]

    return matches


def filter_replacements(original: str, replacements: list[str], vocab: Vocabulary) -> list[str]:
    """
    Heuristic filtering of suggested replacements.

    Applies a small set of heuristics to remove unlikely replacements before ranking:
    - If `original` is not all upper-case, drop suggestions that are all upper-case; otherwise keep only
      all upper-case suggestions.
    - Tokenize each candidate with the module-level `nlp` pipeline and form a list of token lemmas,
      ignoring tokens listed in the module-level `IGNORE_LEMMATIZATION_TOKENS` (for example, `n't`).
    - Keep a replacement only if every remaining token lemma is present in `vocab`.

    Parameters
    ----------
    original : str
        The original part of the context that is considered for replacement.
    replacements : list[str]
        Candidate replacement strings.
    vocab : Vocabulary
        Vocabulary object that restricts acceptable replacements.

    Returns
    -------
    list[str]
        A new list containing replacements that passed the heuristics. The input `replacements`
        list is not modified.

    Notes
    -----
    - Tokenization and lemmatization are performed with the module-level `nlp` pipeline.
    - Tokens whose text appears in `IGNORE_LEMMATIZATION_TOKENS` are excluded from vocabulary checks.
    - The function is conservative: if no replacements remain after filtering, an empty list is returned.
    """

    if not replacements:
        return []

    if not original:
        return replacements

    is_all_caps = original.upper() == original
    if not is_all_caps:
        # keep only non-all-caps from the replacements
        replacements = [item for item in replacements if item.upper() != item]
    else:
        # keep only all-caps from the replacements
        replacements = [item for item in replacements if item.upper() == item]

    if not replacements:
        return []

    # only keep replacements that match the vocabulary.
    docs = list(nlp.pipe(replacements))
    res = []
    for item, doc in zip(replacements, docs):
        lemmas = [tok.lemma_ for tok in doc if tok.text.lower() not in IGNORE_LEMMATIZATION_TOKENS]
        if all(lemma in vocab for lemma in lemmas):
            res.append(item)
    return res



"""
Class for managing interchangeable vocabulary.
"""

from __future__ import annotations

import inflect
import numpy as np
import os
import random
import toml

from .entities import VOWELS
from .verb_conjugator import VerbConjugator


p = inflect.engine()

# dictionary with fixes for conjugation of the mlconjug3
verbs_past = {
    'borrow': 'borrowed',
    'look for': 'looked for',
}


class Vocabulary:
    vocab_file: str | None  # file with the current vocabulary
    name: str | None
    adjectives: list[str]
    adverbs: list[str]
    nouns: list[str]
    verbs: list[str]
    weekdays: list[str] = [
        "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"
    ]
    seasons = [
        "Spring", "Summer", "Autumn", "Winter"
    ]
    months = [
        "January", "February", "March", "April", "May", "June",
        "July", "August", "September", "October", "November", "December"
    ]
    cities = [
        "Beijing", "Berlin", "Busan", "Daejeon", "London", "New York", "Paris", "Seoul", "Tokyo"
    ]
    countries = [
        "Australia", "Brazil", "Canada", "China", "England", "France", "Germany", "India", "Japan",
        "Korea", "Philippines", "Russia", "Thailand", "Vietnam", "USA"
    ]
    nationalities = [
        "Australian", "Brazilian", "Canadian", "Chinese", "English", "French", "German", "Indian", "Japanese",
        "Korean", "Filipino", "Russian", "Thai", "Vietnamese", "American"
    ]
    person_nouns = [
        "anybody", "boy", "brother", "child", "daughter", "everybody",
        "family", "father", "friend", "girl", "husband", "lady", "man", "mother",
        "nobody", "people", "person", "relative", "sister", "somebody", "son",
        "teacher", "wife", "woman"
    ]
    places = [
        "airport", "apartment", "country", "flat", "home", "hospital", "house", "road", "room",
        "school", "shop", "side", "street", "town", "work", "bathroom"
    ]

    def __init__(self, vocab_file: str | None = None):
        self.vocab_file = vocab_file
        self.name = None
        self.adjectives = []
        self.adverbs = []
        self.nouns = []
        self.verbs = []
        self._load()

    @staticmethod
    def load(vocab_file: str) -> Vocabulary:
        return Vocabulary(vocab_file)

    def _load(self):
        if self.vocab_file is None:
            return

        with open(self.vocab_file, "r") as f:
            vocab_dict = toml.loads(f.read())

        if "import" in vocab_dict:
            import_dir = os.path.split(self.vocab_file)[0]
            import_vocab = Vocabulary.load(os.path.join(import_dir, vocab_dict["import"]))
        else:
            import_vocab = Vocabulary()

        self.name = vocab_dict.get("name")
        self.adjectives = sorted(set(import_vocab.adjectives + (vocab_dict.get("adjectives") or [])))
        self.adverbs = sorted(set(import_vocab.adverbs + (vocab_dict.get("adverbs") or [])))
        self.nouns = sorted(set(import_vocab.nouns + (vocab_dict.get("nouns") or [])))
        self.verbs = sorted(set(import_vocab.verbs + (vocab_dict.get("verbs") or [])))
        self.set_ = set(self.adjectives).union(self.adverbs).union(self.nouns).union(self.verbs)

    def __contains__(self, word: str) -> bool:
        return word.lower() in self.set_

    def random_verb(self) -> str:
        return random.choice(self.verbs)

    def random_verb_singular_3rd(self) -> str:
        return VerbConjugator.present_singular(self.random_verb())

    def random_verb_past(self) -> str:
        return VerbConjugator.past_simple(self.random_verb())

    def random_verb_progressive(self) -> str:
        return VerbConjugator.continuous(self.random_verb())

    def random_noun(self) -> str:
        return random.choice(self.nouns)

    def random_anoun(self) -> str:
        noun = self.random_noun()
        article = "an" if noun[0] in VOWELS else "a"
        return f"{article} {noun}"

    def random_noun_plural(self) -> str:
        noun = random.choice(self.nouns)
        return p.plural_noun(noun)

    def random_noun_non_person(self) -> str:
        res = random.choice(self.nouns)
        while res in Vocabulary.person_nouns:
            res = random.choice(self.nouns)
        return res

    def random_anoun_non_person(self) -> str:
        noun = self.random_noun_non_person()
        article = "an" if noun[0] in VOWELS else "a"
        return f"{article} {noun}"

    def random_person(self) -> str:
        res = random.choice(self.person_nouns)
        return res

    def random_adjective(self) -> str:
        return random.choice(self.adjectives)

    def random_adverb(self) -> str:
        return random.choice(self.adverbs)

    def random_weekday(self) -> str:
        return random.choice(self.weekdays)

    def random_season(self) -> str:
        return random.choice(self.seasons)

    def random_month(self) -> str:
        return random.choice(self.months)

    def random_number(self, max) -> str:
        return str(random.randint(0, max))

    def random_number_large(self) -> str:
        # ranges:
        # 1 -- 5k -- 5M -- regular prices
        # 2 -- 3M -- 50M -- car price
        # 3 -- 20M -- 5B -- housing range (loan, purchase)
        # 4 -- 50M -- 10B -- small business related
        # 5 -- 1B -- 100B -- medium business related
        category = np.random.choice([1, 2, 3, 4, 5], p=[0.5, 0.2, 0.15, 0.1, 0.05])
        mins = {
            1: 5000,
            2: 3 * 10**6,
            3: 20 * 10 ** 6,
            4: 50 * 10**6,
            5: 10**9,
        }
        maxes = {
            1: 5 * 10**6,
            2: 50 * 10**6,
            3: 5 * 10**9,
            4: 10 * 10**9,
            5: 100 * 10**9
        }
        res = random.randint(mins[category], maxes[category])
        return f"{res:,}"

    def random_place(self) -> str:
        return random.choice(self.places)

    def random_city(self) -> str:
        return random.choice(self.cities)

    def random_country(self) -> str:
        return random.choice(self.countries)

    def random_nationality(self) -> str:
        return random.choice(self.nationalities)


def load_minilex() -> Vocabulary:
    cur_dir = os.path.split(__file__)[0]
    return Vocabulary.load(os.path.join(cur_dir, "vocabs", "minilex.toml"))


def load_nanolex() -> Vocabulary:
    cur_dir = os.path.split(__file__)[0]
    return Vocabulary.load(os.path.join(cur_dir, "vocabs", "nanolex.toml"))



from __future__ import annotations


# Dictionary with irregular verbs. Each entry contains an array with elements [past simple, past perfect].
# Source: https://github.com/WithEnglishWeCan/generated-english-irregular-verbs/blob/master/irregular.verbs.build.json
#   the first word was taken
irregular_verbs = {
    "abide": ["abode", "abode"],
    "aby": ["abought", "abought"],
    "alight": ["alit", "alit"],
    "arise": ["arose", "arisen"],
    "awake": ["awoke", "awoken"],
    "backbite": ["backbit", "backbitten"],
    "backfit": ["backfit", "backfit"],
    "backlight": ["backlit", "backlit"],
    "backslide": ["backslid", "backslid"],
    "be": [("was", "were"), "been"],
    "bear": ["bore", "born"],
    "beat": ["beat", "beaten"],
    "become": ["became", "become"],
    "bedight": ["bedight", "bedight"],
    "befall": ["befell", "befallen"],
    "beget": ["begot", "begotten"],
    "begin": ["began", "begun"],
    "begird": ["begirt", "begirt"],
    "behight": ["behight", "behight"],
    "behold": ["beheld", "beheld"],
    "belay": ["belayed", "belayed"],
    "bend": ["bent", "bent"],
    "bereave": ["bereft", "bereft"],
    "beseech": ["besought", "besought"],
    "beset": ["beset", "beset"],
    "bespeak": ["bespoke", "bespoken"],
    "bestrew": ["bestrewed", "bestrewed"],
    "bestride": ["bestrode", "bestridden"],
    "bet": ["bet", "bet"],
    "betake": ["betook", "betaken"],
    "bethink": ["bethought", "bethought"],
    "beweep": ["bewept", "bewept"],
    "bid": ["bade", "bidden"],
    "bide": ["bided", "bided"],
    "bind": ["bound", "bound"],
    "bite": ["bit", "bitten"],
    "blaw": ["blawed", "blawn"],
    "bleed": ["bled", "bled"],
    "blend": ["blended", "blended"],
    "bless": ["blessed", "blessed"],
    "blow": ["blew", "blown"],
    "bottle-feed": ["bottle-fed", "bottle-fed"],
    "bowstring": ["bowstrung", "bowstrung"],
    "break": ["broke", "broken"],
    "breast-feed": ["breast-fed", "breast-fed"],
    "breed": ["bred", "bred"],
    "bring": ["brought", "brought"],
    "broadcast": ["broadcast", "broadcast"],
    "browbeat": ["browbeat", "browbeat"],
    "build": ["built", "built"],
    "burn": ["burnt", "burnt"],
    "burst": ["burst", "burst"],
    "bust": ["bust", "bust"],
    "buy": ["bought", "bought"],
    "cast": ["cast", "cast"],
    "catch": ["caught", "caught"],
    "cheerlead": ["cheerled", "cheerled"],
    "chide": ["chid", "chid"],
    "choose": ["chose", "chosen"],
    "clap": ["clapped", "clapped"],
    "clear-cut": ["clear-cut", "clear-cut"],
    "cleave": ["cleft", "cleft"],
    "cleek": ["claught", "cleeked"],
    "clepe": ["cleped", "cleped"],
    "cling": ["clung", "clung"],
    "clothe": ["clad", "clad"],
    "colorbreed": ["colorbred", "colorbred"],
    "colorcast": ["colorcast", "colorcast"],
    "come": ["came", "come"],
    "cost": ["cost", "cost"],
    "cost-cut": ["cost-cut", "cost-cut"],
    "counterdraw": ["counterdrew", "counterdrawn"],
    "counterlight": ["counterlit", "counterlit"],
    "crash-dive": ["crash-dived", "crash-dived"],
    "creep": ["crept", "crept"],
    "creep-feed": ["creep-fed", "creep-fed"],
    "crib-bite": ["crib-bit", "crib-bitten"],
    "cross-bite": ["cross-bit", "cross-bitten"],
    "cross-string": ["cross-strung", "cross-strung"],
    "crossbreed": ["crossbred", "crossbred"],
    "crosscut": ["crosscut", "crosscut"],
    "crosslight": ["crosslit", "crosslit"],
    "crow": ["crowed", "crowed"],
    "cut": ["cut", "cut"],
    "dare": ["dared", "dared"],
    "daydream": ["daydreamt", "daydreamt"],
    "deal": ["dealt", "dealt"],
    "deep-freeze": ["deep-froze", "deep-frozen"],
    "dig": ["dug", "dug"],
    "dight": ["dight", "dight"],
    "dispread": ["dispread", "dispread"],
    "disprove": ["disproved", "disproven"],
    "dive": ["dived", "dived"],
    "do": ["did", "done"],
    "dogfight": ["dogfought", "dogfought"],
    "dow": ["dought", "dought"],
    "draw": ["drew", "drawn"],
    "dream": ["dreamt", "dreamt"],
    "drink": ["drank", "drunk"],
    "drip-feed": ["drip-fed", "drip-fed"],
    "drive": ["drove", "driven"],
    "dwell": ["dwelt", "dwelt"],
    "eat": ["ate", "eaten"],
    "end-run": ["end-ran", "end-run"],
    "engird": ["engirt", "engirt"],
    "engrave": ["engraved", "engraved"],
    "enwind": ["enwound", "enwound"],
    "fall": ["fell", "fallen"],
    "fast-cut": ["fast-cut", "fast-cut"],
    "feed": ["fed", "fed"],
    "feel": ["felt", "felt"],
    "fight": ["fought", "fought"],
    "find": ["found", "found"],
    "fit": ["fit", "fit"],
    "flash-freeze": ["flash-froze", "flash-frozen"],
    "flee": ["fled", "fled"],
    "fling": ["flung", "flung"],
    "floodlight": ["floodlit", "floodlit"],
    "fly": ["flew", "flown"],
    "flyblow": ["flyblew", "flyblown"],
    "forbear": ["forbore", "forborne"],
    "forbid": ["forbade", "forbidden"],
    "force-feed": ["force-fed", "force-fed"],
    "fordo": ["fordid", "fordone"],
    "forecast": ["forecast", "forecast"],
    "forefeel": ["forefelt", "forefelt"],
    "forego": ["forewent", "foregone"],
    "foreknow": ["foreknew", "foreknown"],
    "forerun": ["foreran", "forerun"],
    "foresee": ["foresaw", "foreseen"],
    "foreshow": ["foreshowed", "foreshown"],
    "forespeak": ["forespoke", "forespoken"],
    "foretell": ["foretold", "foretold"],
    "forget": ["forgot", "forgotten"],
    "forgive": ["forgave", "forgiven"],
    "forlese": ["forlore", "forlorn"],
    "forsake": ["forsook", "forsaken"],
    "forswear": ["forswore", "forsworn"],
    "fraught": ["fraught", "fraught"],
    "free-fall": ["free-fell", "free-fallen"],
    "freeze": ["froze", "frozen"],
    "frostbite": ["frostbit", "frostbitten"],
    "gainsay": ["gainsaid", "gainsaid"],
    "geld": ["gelded", "gelded"],
    "get": ["got", "got"],
    "ghostwrite": ["ghostwrote", "ghostwritten"],
    "gild": ["gilt", "gilt"],
    "gin": ["gan", "gan"],
    "gird": ["girded", "girded"],
    "give": ["gave", "given"],
    "gnaw": ["gnawed", "gnawed"],
    "go": ["went", "gone"],
    "grave": ["graved", "graven"],
    "grind": ["ground", "ground"],
    "grow": ["grew", "grown"],
    "hagride": ["hagrode", "hagridden"],
    "halterbreak": ["halterbroke", "halterbroken"],
    "hamstring": ["hamstrung", "hamstrung"],
    "hand-feed": ["hand-fed", "hand-fed"],
    "hand-ride": ["hand-rode", "hand-ridden"],
    "handsew": ["handsewed", "handsewn"],
    "handwrite": ["handwrote", "handwritten"],
    "hang": ["hung", "hung"],
    "have": ["had", "had"],
    "hear": ["heard", "heard"],
    "heave": ["heaved", "heaved"],
    "hew": ["hewed", "hewn"],
    "hide": ["hid", "hidden"],
    "hight": ["hote", "hoten"],
    "hit": ["hit", "hit"],
    "hoise": ["hoised", "hoised"],
    "hoist": ["hoist", "hoist"],
    "hold": ["held", "held"],
    "house-sit": ["house-sat", "house-sat"],
    "housebreak": ["housebroke", "housebroken"],
    "hurt": ["hurt", "hurt"],
    "inbreed": ["inbred", "inbred"],
    "indwell": ["indwelt", "indwelt"],
    "inhold": ["inheld", "inheld"],
    "inlay": ["inlaid", "inlaid"],
    "input": ["input", "input"],
    "inset": ["inset", "inset"],
    "interblend": ["interblended", "interblended"],
    "interbreed": ["interbred", "interbred"],
    "intercut": ["intercut", "intercut"],
    "intergrave": ["intergraved", "intergraved"],
    "interlay": ["interlaid", "interlaid"],
    "interset": ["interset", "interset"],
    "interweave": ["interwove", "interwoven"],
    "interwind": ["interwound", "interwound"],
    "inweave": ["inwove", "inwoven"],
    "inwind": ["inwound", "inwound"],
    "jerry-build": ["jerry-built", "jerry-built"],
    "keep": ["kept", "kept"],
    "ken": ["kent", "kent"],
    "kneel": ["knelt", "knelt"],
    "knit": ["knit", "knit"],
    "know": ["knew", "known"],
    "lade": ["laded", "laden"],
    "landslide": ["landslid", "landslid"],
    "lay": ["laid", "laid"],
    "lead": ["led", "led"],
    "lean": ["leant", "leant"],
    "leap": ["leapt", "leapt"],
    "learn": ["learnt", "learnt"],
    "leave": ["left", "left"],
    "lend": ["lent", "lent"],
    "lep": ["lept", "lept"],
    "let": ["let", "let"],
    "lie": ["lay", "lain"],
    "light": ["lit", "lit"],
    "linebreed": ["linebred", "linebred"],
    "lip-read": ["lip-read", "lip-read"],
    "lose": ["lost", "lost"],
    "make": ["made", "made"],
    "mean": ["meant", "meant"],
    "meet": ["met", "met"],
    "melt": ["melted", "molten"],
    "misbecome": ["misbecame", "misbecome"],
    "miscast": ["miscast", "miscast"],
    "mischoose": ["mischose", "mischosen"],
    "miscut": ["miscut", "miscut"],
    "misdeal": ["misdealt", "misdealt"],
    "misdo": ["misdid", "misdone"],
    "misfall": ["misfell", "misfallen"],
    "misfeed": ["misfed", "misfed"],
    "misgive": ["misgave", "misgiven"],
    "mishear": ["misheard", "misheard"],
    "mishit": ["mishit", "mishit"],
    "misknow": ["misknew", "misknown"],
    "mislay": ["mislaid", "mislaid"],
    "mislead": ["misled", "misled"],
    "mislearn": ["mislearnt", "mislearnt"],
    "misread": ["misread", "misread"],
    "missay": ["missaid", "missaid"],
    "missend": ["missent", "missent"],
    "misset": ["misset", "misset"],
    "misspeak": ["misspoke", "misspoken"],
    "misspell": ["misspelt", "misspelt"],
    "misspend": ["misspent", "misspent"],
    "misswear": ["misswore", "missworn"],
    "mistake": ["mistook", "mistaken"],
    "misteach": ["mistaught", "mistaught"],
    "mistell": ["mistold", "mistold"],
    "misthink": ["misthought", "misthought"],
    "misunderstand": ["misunderstood", "misunderstood"],
    "miswear": ["misswore", "misworn"],
    "miswed": ["miswed", "miswed"],
    "miswrite": ["miswrote", "miswritten"],
    "mow": ["mowed", "mown"],
    "naysay": ["naysaid", "naysaid"],
    "nose-dive": ["nose-dived", "nose-dived"],
    "offset": ["offset", "offset"],
    "outbid": ["outbid", "outbid"],
    "outbreed": ["outbred", "outbred"],
    "outdo": ["outdid", "outdone"],
    "outdraw": ["outdrew", "outdrawn"],
    "outdrink": ["outdrank", "outdrunk"],
    "outdrive": ["outdrove", "outdriven"],
    "outfight": ["outfought", "outfought"],
    "outfly": ["outflew", "outflown"],
    "outgrow": ["outgrew", "outgrown"],
    "outlay": ["outlaid", "outlaid"],
    "outleap": ["outleapt", "outleapt"],
    "output": ["output", "output"],
    "outride": ["outrode", "outridden"],
    "outrun": ["outran", "outrun"],
    "outsee": ["outsaw", "outseen"],
    "outsell": ["outsold", "outsold"],
    "outshine": ["outshone", "outshone"],
    "outshoot": ["outshot", "outshot"],
    "outsing": ["outsang", "outsung"],
    "outsit": ["outsat", "outsat"],
    "outsleep": ["outslept", "outslept"],
    "outsmell": ["outsmelled", "outsmelled"],
    "outspeak": ["outspoke", "outspoken"],
    "outspeed": ["outsped", "outsped"],
    "outspend": ["outspent", "outspent"],
    "outspin": ["outspun", "outspun"],
    "outspring": ["outsprang", "outsprung"],
    "outstand": ["outstood", "outstood"],
    "outswear": ["outswore", "outsworn"],
    "outswim": ["outswam", "outswum"],
    "outtell": ["outtold", "outtold"],
    "outthink": ["outthought", "outthought"],
    "outthrow": ["outthrew", "outthrown"],
    "outwear": ["outwore", "outworn"],
    "outwind": ["outwound", "outwound"],
    "outwork": ["outworked", "outworked"],
    "outwrite": ["outwrote", "outwritten"],
    "overbear": ["overbore", "overborne"],
    "overbid": ["overbid", "overbid"],
    "overblow": ["overblew", "overblown"],
    "overbreed": ["overbred", "overbred"],
    "overbuild": ["overbuilt", "overbuilt"],
    "overbuy": ["overbought", "overbought"],
    "overcast": ["overcast", "overcast"],
    "overcome": ["overcame", "overcome"],
    "overcut": ["overcut", "overcut"],
    "overdo": ["overdid", "overdone"],
    "overdraw": ["overdrew", "overdrawn"],
    "overdrink": ["overdrank", "overdrunk"],
    "overdrive": ["overdrove", "overdriven"],
    "overeat": ["overate", "overeaten"],
    "overfeed": ["overfed", "overfed"],
    "overfly": ["overflew", "overflown"],
    "overgrow": ["overgrew", "overgrown"],
    "overhang": ["overhung", "overhung"],
    "overhear": ["overheard", "overheard"],
    "overlay": ["overlaid", "overlaid"],
    "overleap": ["overleapt", "overleapt"],
    "overlearn": ["overlearnt", "overlearnt"],
    "overset": ["overset", "overset"],
    "oversew": ["oversewed", "oversewn"],
    "overshoot": ["overshot", "overshot"],
    "oversleep": ["overslept", "overslept"],
    "oversow": ["oversowed", "oversowed"],
    "overspeak": ["overspoke", "overspoken"],
    "overspend": ["overspent", "overspent"],
    "overspill": ["overspilt", "overspilt"],
    "overspin": ["overspun", "overspun"],
    "overspread": ["overspread", "overspread"],
    "overspring": ["oversprang", "oversprung"],
    "overstand": ["overstood", "overstood"],
    "overstrew": ["overstrewed", "overstrewed"],
    "overstride": ["overstrode", "overstridden"],
    "overstrike": ["overstruck", "overstruck"],
    "overstring": ["overstrung", "overstrung"],
    "overtake": ["overtook", "overtaken"],
    "overthink": ["overthought", "overthought"],
    "overthrow": ["overthrew", "overthrown"],
    "overwear": ["overwore", "overworn"],
    "overwind": ["overwound", "overwound"],
    "overwithhold": ["overwithheld", "overwithheld"],
    "overwrite": ["overwrote", "overwritten"],
    "partake": ["partook", "partaken"],
    "pay": ["paid", "paid"],
    "pen": ["pent", "pent"],
    "pinch-hit": ["pinch-hit", "pinch-hit"],
    "plead": ["pled", "pled"],
    "potshot": ["potshot", "potshot"],
    "practice-teach": ["practice-taught", "practice-taught"],
    "prebind": ["prebound", "prebound"],
    "prebuild": ["prebuilt", "prebuilt"],
    "precut": ["precut", "precut"],
    "predo": ["predid", "predone"],
    "premake": ["premade", "premade"],
    "prepay": ["prepaid", "prepaid"],
    "presell": ["presold", "presold"],
    "preset": ["preset", "preset"],
    "preshrink": ["preshrank", "preshrunk"],
    "presplit": ["presplit", "presplit"],
    "price-cut": ["price-cut", "price-cut"],
    "proofread": ["proofread", "proofread"],
    "prove": ["proved", "proven"],
    "put": ["put", "put"],
    "quick-freeze": ["quick-froze", "quick-frozen"],
    "quit": ["quit", "quit"],
    "rap": ["rapped", "rapped"],
    "re-prove": ["re-proved", "re-proved"],
    "read": ["read", "read"],
    "reave": ["reft", "reft"],
    "reawake": ["reawoke", "reawoken"],
    "rebid": ["rebid", "rebid"],
    "rebind": ["rebound", "rebound"],
    "rebroadcast": ["rebroadcast", "rebroadcast"],
    "rebuild": ["rebuilt", "rebuilt"],
    "recast": ["recast", "recast"],
    "recut": ["recut", "recut"],
    "redd": ["redd", "redd"],
    "redeal": ["redealt", "redealt"],
    "redo": ["redid", "redone"],
    "redraw": ["redrew", "redrawn"],
    "reeve": ["rove", "rove"],
    "refit": ["refit", "refit"],
    "regrind": ["reground", "reground"],
    "regrow": ["regrew", "regrown"],
    "rehang": ["rehung", "rehung"],
    "rehear": ["reheard", "reheard"],
    "reknit": ["reknit", "reknit"],
    "relay": ["relaid", "relaid"],
    "relearn": ["relearnt", "relearnt"],
    "relight": ["relit", "relit"],
    "remake": ["remade", "remade"],
    "rend": ["rent", "rent"],
    "repay": ["repaid", "repaid"],
    "reread": ["reread", "reread"],
    "rerun": ["reran", "rerun"],
    "resell": ["resold", "resold"],
    "resend": ["resent", "resent"],
    "reset": ["reset", "reset"],
    "resew": ["resewed", "resewn"],
    "reshoot": ["reshot", "reshot"],
    "resit": ["resat", "resat"],
    "resow": ["resowed", "resown"],
    "respell": ["respelled", "respelled"],
    "restring": ["restrung", "restrung"],
    "retake": ["retook", "retaken"],
    "reteach": ["retaught", "retaught"],
    "retear": ["retore", "retorn"],
    "retell": ["retold", "retold"],
    "rethink": ["rethought", "rethought"],
    "retread": ["retrod", "retrodden"],
    "retrofit": ["retrofit", "retrofit"],
    "rewake": ["rewoke", "rewoken"],
    "reweave": ["rewove", "rewoven"],
    "rewed": ["rewed", "rewed"],
    "rewet": ["rewet", "rewet"],
    "rewin": ["rewon", "rewon"],
    "rewind": ["rewound", "rewound"],
    "rewrite": ["rewrote", "rewritten"],
    "rid": ["rid", "rid"],
    "ride": ["rode", "ridden"],
    "ring": ["rang", "rung"],
    "rise": ["rose", "risen"],
    "rive": ["rived", "riven"],
    "rough-hew": ["rough-hewed", "rough-hewn"],
    "roughcast": ["roughcast", "roughcast"],
    "run": ["ran", "run"],
    "sand-cast": ["sand-cast", "sand-cast"],
    "saw": ["sawed", "sawn"],
    "say": ["said", "said"],
    "see": ["saw", "seen"],
    "seek": ["sought", "sought"],
    "self-feed": ["self-fed", "self-fed"],
    "self-sow": ["self-sowed", "self-sown"],
    "sell": ["sold", "sold"],
    "send": ["sent", "sent"],
    "set": ["set", "set"],
    "sew": ["sewed", "sewn"],
    "shake": ["shook", "shaken"],
    "shave": ["shaved", "shaven"],
    "shear": ["shore", "shorn"],
    "shed": ["shed", "shed"],
    "shend": ["shent", "shent"],
    "shew": ["shewed", "shewn"],
    "shine": ["shone", "shone"],
    "shit": ["shat", "shat"],
    "shoe": ["shod", "shod"],
    "shoot": ["shot", "shot"],
    "shortcut": ["shortcut", "shortcut"],
    "show": ["showed", "shown"],
    "shrink": ["shrank", "shrunk"],
    "shrive": ["shrove", "shriven"],
    "shut": ["shut", "shut"],
    "sidewind": ["sidewound", "sidewound"],
    "sight-read": ["sight-read", "sight-read"],
    "sightsee": ["sightsaw", "sightseen"],
    "sing": ["sang", "sung"],
    "sink": ["sank", "sunk"],
    "sit": ["sat", "sat"],
    "skywrite": ["skywrote", "skywritten"],
    "slay": ["slew", "slain"],
    "sleep": ["slept", "slept"],
    "slide": ["slid", "slid"],
    "sling": ["slung", "slung"],
    "slink": ["slunk", "slunk"],
    "slit": ["slit", "slit"],
    "smell": ["smelt", "smelt"],
    "smite": ["smote", "smitten"],
    "snapshoot": ["snapshot", "snapshot"],
    "sneak": ["sneaked", "sneaked"],
    "soothsay": ["soothsaid", "soothsaid"],
    "sow": ["sowed", "sown"],
    "speak": ["spoke", "spoken"],
    "speed": ["sped", "sped"],
    "spell": ["spelt", "spelt"],
    "spellbind": ["spellbound", "spellbound"],
    "spend": ["spent", "spent"],
    "spill": ["spilt", "spilt"],
    "spin": ["span", "spun"],
    "spit": ["spat", "spat"],
    "split": ["split", "split"],
    "spoil": ["spoilt", "spoilt"],
    "spoonfeed": ["spoonfed", "spoonfed"],
    "spread": ["spread", "spread"],
    "spring": ["sprang", "sprung"],
    "stall-feed": ["stall-fed", "stall-fed"],
    "stand": ["stood", "stood"],
    "stave": ["stove", "stove"],
    "steal": ["stole", "stolen"],
    "stick": ["stuck", "stuck"],
    "sting": ["stung", "stung"],
    "stink": ["stank", "stunk"],
    "straphang": ["straphung", "straphung"],
    "strew": ["strewed", "strewn"],
    "stride": ["strode", "stridden"],
    "strike": ["struck", "struck"],
    "string": ["strung", "strung"],
    "strip": ["stript", "stript"],
    "strive": ["strove", "striven"],
    "sublet": ["sublet", "sublet"],
    "sunburn": ["sunburned", "sunburned"],
    "swear": ["swore", "sworn"],
    "sweat": ["sweat", "sweat"],
    "sweep": ["swept", "swept"],
    "swell": ["swelled", "swollen"],
    "swim": ["swam", "swum"],
    "swing": ["swung", "swung"],
    "swink": ["swank", "swonken"],
    "switch-hit": ["switch-hit", "switch-hit"],
    "take": ["took", "taken"],
    "teach": ["taught", "taught"],
    "tear": ["tore", "torn"],
    "telecast": ["telecast", "telecast"],
    "tell": ["told", "told"],
    "test-drive": ["test-drove", "test-driven"],
    "test-fly": ["test-flew", "test-flown"],
    "think": ["thought", "thought"],
    "thrive": ["throve", "thriven"],
    "throw": ["threw", "thrown"],
    "thrust": ["thrust", "thrust"],
    "thunderstrike": ["thunderstruck", "thunderstruck"],
    "tine": ["tint", "tint"],
    "tread": ["trod", "trodden"],
    "troubleshoot": ["troubleshot", "troubleshot"],
    "typecast": ["typecast", "typecast"],
    "typeset": ["typeset", "typeset"],
    "typewrite": ["typewrote", "typewritten"],
    "unbear": ["unbore", "unborn"],
    "unbend": ["unbent", "unbent"],
    "unbind": ["unbound", "unbound"],
    "unbuild": ["unbuilt", "unbuilt"],
    "unclothe": ["unclad", "unclad"],
    "underbid": ["underbid", "underbid"],
    "underbuy": ["underbought", "underbought"],
    "undercut": ["undercut", "undercut"],
    "underdo": ["underdid", "underdone"],
    "underfeed": ["underfed", "underfed"],
    "undergird": ["undergirt", "undergirt"],
    "undergo": ["underwent", "undergone"],
    "underlay": ["underlaid", "underlaid"],
    "underlet": ["underlet", "underlet"],
    "underlie": ["underlay", "underlain"],
    "underpay": ["underpaid", "underpaid"],
    "underrun": ["underran", "underrun"],
    "undersell": ["undersold", "undersold"],
    "undershoot": ["undershot", "undershot"],
    "underspend": ["underspent", "underspent"],
    "understand": ["understood", "understood"],
    "undertake": ["undertook", "undertaken"],
    "underthrust": ["underthrust", "underthrust"],
    "underwrite": ["underwrote", "underwritten"],
    "undo": ["undid", "undone"],
    "undraw": ["undrew", "undrawn"],
    "unfreeze": ["unfroze", "unfrozen"],
    "unhang": ["unhung", "unhung"],
    "unhide": ["unhid", "unhidden"],
    "unhold": ["unheld", "unheld"],
    "unknit": ["unknit", "unknit"],
    "unlade": ["unladed", "unladen"],
    "unlay": ["unlaid", "unlaid"],
    "unlearn": ["unlearnt", "unlearnt"],
    "unmake": ["unmade", "unmade"],
    "unreeve": ["unrove", "unrove"],
    "unsay": ["unsaid", "unsaid"],
    "unsell": ["unsold", "unsold"],
    "unsew": ["unsewed", "unsewn"],
    "unsling": ["unslung", "unslung"],
    "unspeak": ["unspoke", "unspoken"],
    "unspin": ["unspun", "unspun"],
    "unstick": ["unstuck", "unstuck"],
    "unstring": ["unstrung", "unstrung"],
    "unswear": ["unswore", "unsworn"],
    "unteach": ["untaught", "untaught"],
    "unthink": ["unthought", "unthought"],
    "untread": ["untrod", "untrodden"],
    "unweave": ["unwove", "unwoven"],
    "unwind": ["unwound", "unwound"],
    "unwrite": ["unwrote", "unwritten"],
    "upbuild": ["upbuilt", "upbuilt"],
    "uphold": ["upheld", "upheld"],
    "uppercut": ["uppercut", "uppercut"],
    "uprise": ["uprose", "uprisen"],
    "upset": ["upset", "upset"],
    "upspring": ["upsprang", "upsprung"],
    "upsweep": ["upswept", "upswept"],
    "uptear": ["uptore", "uptorn"],
    "vex": ["vext", "vext"],
    "wake": ["woke", "woken"],
    "waylay": ["waylaid", "waylaid"],
    "wear": ["wore", "worn"],
    "weave": ["wove", "woven"],
    "wed": ["wed", "wed"],
    "weep": ["wept", "wept"],
    "wend": ["wended", "wended"],
    "wet": ["wet", "wet"],
    "whipsaw": ["whipsawed", "whipsawn"],
    "win": ["won", "won"],
    "wind": ["wound", "wound"],
    "winterfeed": ["winterfed", "winterfed"],
    "wiredraw": ["wiredrew", "wiredrawn"],
    "wit": ["wist", "wist"],
    "withdraw": ["withdrew", "withdrawn"],
    "withhold": ["withheld", "withheld"],
    "withstand": ["withstood", "withstood"],
    "wont": ["wont", "wont"],
    "work": ["worked", "worked"],
    "wrap": ["wrapped", "wrapped"],
    "wreak": ["wreaked", "wreaked"],
    "wring": ["wrung", "wrung"],
    "write": ["wrote", "written"],
    "zinc": ["zinced", "zinced"]
}


class VerbConjugator:
    @staticmethod
    def past_simple(verb: str):
        if verb in irregular_verbs:
            return irregular_verbs[verb][0]

        exceptions = {
            "cry": "cried",
            "drop": "dropped",
            "look for": "looked for",
            "stop": "stopped",
            "travel": "travelled",
        }
        if verb in exceptions:
            return exceptions[verb]

        if verb.endswith("e"):
            return verb + "d"

        return verb + "ed"

    @staticmethod
    def present_singular(verb: str):
        exceptions = {
            "cry": "cries",
            "do": "goes",
            "go": "goes",
            "have": "has",
            "look for": "looks for",
            "teach": "teaches",
            "worry": "worries",
        }
        if verb in exceptions:
            return exceptions[verb]

        return verb + "s"

    @staticmethod
    def continuous(verb: str):
        exceptions = {
            "begin": "beginning",
            "let": "letting",
            "look for": "looking for",
            "wait": "waiting",
        }
        if verb in exceptions:
            return exceptions[verb]

        if verb.endswith("e"):
            return verb[:-1] + "ing"

        if verb.endswith("t"):
            return verb + "ting"

        return verb + "ing"



from .exercise_generator import ExerciseGenerator
from .grammar_checker import GrammarChecker
from .registry import GrammarRegistry
from .reranker import Reranker
from .vocabulary import Vocabulary, load_minilex
